try:
    import unittest2 as unittest
    unittest # pragma no cover stfu pyflakes
except ImportError:  # pragma no cover
    import unittest

import churro
import transaction


class ChurroTests(unittest.TestCase):

    def setUp(self):
        import tempfile
        self.tmp = tempfile.mkdtemp('.churro-test')

    def tearDown(self):
        import shutil
        transaction.abort()
        shutil.rmtree(self.tmp)

    def make_one(self, **kw):
        from churro import Churro as test_class
        return test_class(self.tmp, **kw)

    def test_empty_repo(self):
        repo = self.make_one()
        folder = repo.root()
        self.assertEqual(len(folder), 0)

    def test_save_and_retrieve_one_object(self):
        repo = self.make_one()
        root = repo.root()
        self.assertEqual(repo.root(), root)
        obj = TestClass('foo', 'bar')
        root['test'] = obj
        self.assertIs(root['test'], obj)
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        obj = root['test']
        self.assertEqual(obj.one, 'foo')
        self.assertEqual(obj.two, 'bar')
        obj.two = 'dos'
        self.assertEqual(obj.two, 'dos')
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        obj = root['test']
        self.assertEqual(obj.two, 'dos')

    def test_save_and_retrieve_one_object_in_a_folder(self):
        repo = self.make_one()
        root = repo.root()
        folder = TestFolder('uno', 'dos')
        self.assertEqual(len(folder), 0)
        root['folder'] = folder
        self.assertEqual(len(folder), 0)
        obj = TestClass('foo', 'bar')
        folder['test'] = obj
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        folder = root['folder']
        self.assertEqual(folder.one, 'uno')
        self.assertEqual(folder.two, 'dos')
        obj = folder['test']
        self.assertEqual(obj.one, 'foo')
        self.assertEqual(obj.two, 'bar')
        folder.one = 'un'
        obj.two = 'deux'
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        folder = root['folder']
        self.assertEqual(folder.one, 'un')
        obj = folder['test']
        self.assertEqual(obj.two, 'deux')

    def test_folder_ops(self):
        repo = self.make_one()
        root = repo.root()
        self.assertFalse(bool(root))
        root['a'] = a = TestClass('a', 'b')
        root['b'] = b = TestClass('c', 'd')

        self.assertEqual(len(root), 2)
        self.assertEqual(sorted(root.keys()), ['a', 'b'])
        self.assertEqual(sorted(iter(root)), ['a', 'b'])
        self.assertEqual(list(root.values()), [a, b])
        self.assertEqual(list(root.items()), [('a', a), ('b', b)])
        self.assertTrue(bool(root))
        self.assertIn('a', root)

        transaction.commit()
        repo = self.make_one()
        root = repo.root()

        self.assertEqual(len(root), 2)
        self.assertEqual(list(root.values()), [root['a'], root['b']])
        self.assertEqual(list(root.items()), [
            ('a', root['a']), ('b', root['b'])])
        self.assertEqual(sorted(root.keys()), ['a', 'b'])
        self.assertEqual(list(iter(root)), ['a', 'b'])
        self.assertTrue(bool(root))
        self.assertIn('a', root)

        root['c'] = TestFolder('e', 'f')
        self.assertEqual(len(root['c']), 0)

        with self.assertRaises(KeyError):
            root['d']

    def test_delete(self):
        repo = self.make_one()
        root = repo.root()
        root['a'] = TestClass('a', 'b')
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        self.assertIn('a', root)
        del root['a']
        self.assertNotIn('a', root)
        with self.assertRaises(KeyError):
            del root['a']
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        self.assertNotIn('a', root)
        with self.assertRaises(KeyError):
            del root['a']

    def test_pop(self):
        repo = self.make_one()
        root = repo.root()
        root['a'] = TestClass('a', 'b')
        root['b'] = TestClass('c', 'd')
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        self.assertIn('a', root)
        a = root['a']
        self.assertIs(root.pop('a'), a)
        self.assertNotIn('a', root)
        with self.assertRaises(KeyError):
            root.pop('a')
        self.assertEqual(root.pop('a', None), None)
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        self.assertNotIn('a', root)
        with self.assertRaises(KeyError):
            root.pop('a')
        self.assertEqual(root.pop('a', None), None)
        self.assertEqual(root.pop('b').one, 'c')

    def test_remove_folder(self):
        repo = self.make_one()
        root = repo.root()
        root['a'] = TestFolder('a', 'b')
        root['a']['b'] = TestFolder('c', 'd')
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        self.assertIn('a', root)
        del root['a']
        self.assertNotIn('a', root)
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        self.assertNotIn('a', root)
        transaction.commit() # coverage

    def test_deep_structure(self):
        repo = self.make_one()
        root = repo.root()
        obj = TestClass('foo', TestClass('bar', 'baz'))
        root['uhoh'] = obj
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        obj = root['uhoh']
        self.assertEqual(obj.one, 'foo')
        self.assertEqual(obj.two.one, 'bar')
        self.assertEqual(obj.two.two, 'baz')
        obj.two.one = 'bozo'
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        obj = root['uhoh']
        self.assertEqual(obj.one, 'foo')
        self.assertEqual(obj.two.one, 'bozo')
        self.assertEqual(obj.two.two, 'baz')

    def test_not_serializable(self):
        repo = self.make_one()
        root = repo.root()
        obj = TestClass('foo', NotSerializable())
        root['oops'] = obj
        with self.assertRaises(ValueError):
            transaction.commit()

    def test_save_and_retrieve_one_object_with_dates(self):
        import datetime
        dob = datetime.date(1975, 7, 7)
        dtob = datetime.datetime(2010, 5, 12, 2, 42)
        repo = self.make_one()
        root = repo.root()
        self.assertEqual(repo.root(), root)
        obj = TestClassWithDateProperties(dob, dtob)
        root['test'] = obj
        self.assertIs(root['test'], obj)
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        obj = root['test']
        self.assertEqual(obj.three, dob)
        self.assertEqual(obj.four, dtob )
        obj.four = dtob = datetime.datetime(2010, 5, 15, 20, 36)
        self.assertEqual(obj.four, dtob)
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        obj = root['test']
        self.assertEqual(obj.four, dtob)
        obj.three = None
        obj.four = None
        self.assertEqual(obj.three, None)
        self.assertEqual(obj.four, None)
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        obj = root['test']
        self.assertEqual(obj.three, None)
        self.assertEqual(obj.four, None)

        with self.assertRaises(ValueError):
            obj.three = 'foo'
        with self.assertRaises(ValueError):
            obj.four = 'bar'

    def test_persistent_dict_and_list(self):
        repo = self.make_one()
        root = repo.root()
        obj = TestClass(
            churro.PersistentDict({'one': 1, 'two': 2}),
            churro.PersistentList([1, 2]))
        root['test'] = obj
        self.assertEqual(obj.one['one'], 1)
        self.assertEqual(obj.two[0], 1)
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        obj = root['test']
        obj.one['one'] = 'one'
        self.assertEqual(obj.one['one'], 'one')
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        obj = root['test']
        self.assertEqual(obj.one['one'], 'one')
        obj.two[0] = 'one'
        self.assertEqual(obj.two[0], 'one')
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        obj = root['test']
        self.assertEqual(obj.two[0], 'one')

    def test_deep_structure_with_persistent_dict(self):
        repo = self.make_one()
        root = repo.root()
        obj = TestClass(churro.PersistentDict(), None)
        obj.one['foo'] = TestClass('bar', 'baz')
        root['test'] = obj
        self.assertEqual(obj.one['foo'].one, 'bar')
        self.assertEqual(obj.one['foo'].two, 'baz')
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        obj = root['test']
        obj.one['foo'].two = 'bathsalts'
        self.assertEqual(obj.one['foo'].two, 'bathsalts')
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        obj = root['test']
        self.assertEqual(obj.one['foo'].two, 'bathsalts')

    def test_deep_structure_with_persistent_list(self):
        repo = self.make_one()
        root = repo.root()
        obj = TestClass(churro.PersistentList(), None)
        obj.one.append(TestClass('bar', 'baz'))
        root['test'] = obj
        self.assertEqual(obj.one[0].one, 'bar')
        self.assertEqual(obj.one[0].two, 'baz')
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        obj = root['test']
        obj.one[0].two = 'bathsalts'
        self.assertEqual(obj.one[0].two, 'bathsalts')
        transaction.commit()

        repo = self.make_one()
        root = repo.root()
        obj = root['test']
        self.assertEqual(obj.one[0].two, 'bathsalts')


class TestDottedNameResolver(unittest.TestCase):

    def call_fut(self, name):
        from churro import _resolve_dotted_name as fut
        return fut(name)

    def test_it(self):
        C = self.call_fut('email.message.Message')
        from email.message import Message
        self.assertEqual(C, Message)


class CollectionWrappersTestBase(unittest.TestCase):

    def assertNewEq(self, orig, one, two):
        assert one is not orig
        self.assertEqual(one, two)

    def assertMutated(self):
        self.assertTrue(self.o.has_been_mutated)

    def assertNotMutated(self):
        self.assertFalse(self.o.has_been_mutated)


class TestDictWrapper(CollectionWrappersTestBase):

    def setUp(self):
        from churro.collection_wrappers import DictWrapper
        class Derived(DictWrapper):
            has_been_mutated = False
            def mutated(self):
                self.has_been_mutated = True
        self.o = Derived({1: 2, 3: 4})

    def test__contains__(self):
        self.assertTrue(1 in self.o)
        self.assertNotMutated()

    def test__delitem__(self):
        del self.o[1]
        self.assertEqual(self.o, {3: 4})
        self.assertMutated()

    def test__getitem__(self):
        self.assertEqual(self.o[1], 2)
        self.assertNotMutated()

    def test__hash__(self):
        with self.assertRaises(TypeError):
            hash(self.o)
        self.assertNotMutated()

    def test__iter__(self):
        self.assertEqual(sorted(iter(self.o)), [1, 3])
        self.assertNotMutated()

    def test__len__(self):
        self.assertEqual(len(self.o), 2)
        self.assertNotMutated()

    def test__ne__(self):
        self.assertFalse(self.o != {1: 2, 3: 4})
        self.assertNotMutated()

    def test__repr__(self):
        self.assertEqual(repr(self.o), 'Derived({1: 2, 3: 4})')
        self.assertNotMutated()

    def test__setitem__(self):
        self.o[5] = 6
        self.assertEqual(self.o[5], 6)
        self.assertMutated()

    def test_clear(self):
        self.o.clear()
        self.assertEqual(self.o, {})
        self.assertMutated()

    def test_copy(self):
        from churro.collection_wrappers import DictWrapper
        other = self.o.copy()
        self.assertNewEq(self.o, other, self.o)
        self.assertIsInstance(other, DictWrapper)
        self.assertNotMutated()

    def test_get(self):
        self.assertEqual(self.o.get(1), 2)
        self.assertEqual(self.o.get(5), None)
        self.assertEqual(self.o.get(5, 6), 6)
        self.assertNotMutated()

    def test_items(self):
        self.assertEqual(sorted(self.o.items()), [(1, 2), (3, 4)])
        self.assertNotMutated()

    def test_keys(self):
        self.assertEqual(sorted(self.o.keys()), [1, 3])
        self.assertNotMutated()

    def test_pop(self):
        self.assertEqual(self.o.pop(1), 2)
        self.assertEqual(self.o, {3: 4})
        self.assertMutated()

    def test_popitem(self):
        self.assertEqual(self.o.popitem(), (1, 2)) # brittle!
        self.assertEqual(self.o, {3: 4})
        self.assertMutated()

    def test_setdefault(self):
        self.assertEqual(self.o.setdefault(1, 42), 2)
        self.assertEqual(self.o.setdefault(5, 6), 6)
        self.assertEqual(self.o, {1: 2, 3: 4, 5: 6})
        self.assertMutated()

    def test_update(self):
        self.o.update({3: 42, 5: 6})
        self.assertEqual(self.o, {1: 2, 3: 42, 5: 6})
        self.assertMutated()

    def test_values(self):
        self.assertEqual(sorted(self.o.values()), [2, 4])
        self.assertNotMutated()


class TestListWrapper(CollectionWrappersTestBase):

    def setUp(self):
        from churro.collection_wrappers import ListWrapper
        class Derived(ListWrapper):
            has_been_mutated = False
            def mutated(self):
                self.has_been_mutated = True
        self.o = Derived([1, 2, 3, 2, 1])

    def test__contains__(self):
        self.assertTrue(2 in self.o)
        self.assertNotMutated()

    def test__delitem__(self):
        del self.o[1]
        self.assertEqual(self.o, [1, 3, 2, 1])
        self.assertMutated()

    def test__delslice__(self):
        del self.o[1:3]
        self.assertEqual(self.o, [1, 2, 1])
        self.assertMutated()

    def test__getitem__(self):
        self.assertEqual(self.o[1], 2)
        self.assertNotMutated()

    def test__getslice__(self):
        self.assertEqual(self.o[1:3], [2, 3])
        self.assertNotMutated()

    def test__hash__(self):
        with self.assertRaises(TypeError):
            hash(self.o)
        self.assertNotMutated()

    def test__iadd__(self):
        self.o += [4, 5]
        self.assertEqual(self.o, [1, 2, 3, 2, 1, 4, 5])
        self.assertNotMutated()

    def test__imul__(self):
        self.o *= 2
        self.assertEqual(self.o, [1, 2, 3, 2, 1, 1, 2, 3, 2, 1])
        self.assertNotMutated()

    def test__iter__(self):
        self.assertEqual(list(iter(self.o)), [1, 2, 3, 2, 1])
        self.assertNotMutated()

    def test__len__(self):
        self.assertEqual(len(self.o), 5)
        self.assertNotMutated()

    def test__mul__(self):
        self.assertEqual(self.o * 2, [1, 2, 3, 2, 1, 1, 2, 3, 2, 1])
        self.assertEqual(self.o, [1, 2, 3, 2, 1])
        self.assertNotMutated()

    def test__ne__(self):
        self.assertFalse(self.o != [1, 2, 3, 2, 1])
        self.assertNotMutated()

    def test__repr__(self):
        self.assertEqual(repr(self.o), 'Derived([1, 2, 3, 2, 1])')
        self.assertNotMutated()

    def test__rmul__(self):
        self.assertEqual(2 * self.o, [1, 2, 3, 2, 1, 1, 2, 3, 2, 1])
        self.assertEqual(self.o, [1, 2, 3, 2, 1])
        self.assertNotMutated()

    def test__setitem__(self):
        self.o[2] = 6
        self.assertEqual(self.o, [1, 2, 6, 2, 1])
        self.assertMutated()

    def test__setslice__(self):
        self.o[2:4] = [7, 8, 9]
        self.assertEqual(self.o, [1, 2, 7, 8 , 9, 1])
        self.assertMutated()

    def test_append(self):
        self.o.append(6)
        self.assertEqual(self.o, [1, 2, 3, 2, 1, 6])
        self.assertMutated()

    def test_count(self):
        self.assertEqual(self.o.count(2), 2)
        self.assertNotMutated()

    def test_extend(self):
        self.o.extend([6, 7])
        self.assertEqual(self.o, [1, 2, 3, 2, 1, 6, 7])
        self.assertMutated()

    def test_index(self):
        self.assertEqual(self.o.index(3), 2)
        self.assertNotMutated()

    def test_insert(self):
        self.o.insert(0, 0)
        self.assertEqual(self.o, [0, 1, 2, 3, 2, 1])
        self.assertMutated()

    def test_pop(self):
        self.assertEqual(self.o.pop(0), 1)
        self.assertEqual(self.o, [2, 3, 2, 1])
        self.assertMutated()

    def test_remove(self):
        self.o.remove(3)
        self.assertEqual(self.o, [1, 2, 2, 1])
        self.assertMutated()

    def test_reverse(self):
        del self.o[0]
        self.o.reverse()
        self.assertEqual(self.o, [1, 2, 3, 2])
        self.assertMutated()

    def test_sort(self):
        self.o.sort()
        self.assertEqual(self.o, [1, 1, 2, 2, 3])
        self.assertMutated()


class TestClass(churro.Persistent):
    one = churro.PersistentProperty()
    two = churro.PersistentProperty()

    def __init__(self, one, two):
        self.one = one
        self.two = two


class TestClassWithDateProperties(TestClass):
    three = churro.PersistentDate()
    four = churro.PersistentDatetime()

    def __init__(self, three, four):
        self.three = three
        self.four = four


class TestFolder(churro.PersistentFolder, TestClass):
    pass


class NotSerializable(object):
    """Nuh uh, no way."""





class DictWrapper(object):
    def __init__(self, *args):
        self.data = dict(*args)

    def __contains__(self, key):
        return self.data.__contains__(key)

    def __delitem__(self, key):
        self.mutated()
        return self.data.__delitem__(key)

    def __eq__(self, other):
        return self.data.__eq__(self._cast(other))

    def __getitem__(self, key):
        return self.data.__getitem__(key)

    def __hash__(self):
        return self.data.__hash__()

    def __iter__(self):
        return self.data.__iter__()

    def __len__(self):
        return self.data.__len__()

    def __ne__(self, other):
        return self.data.__ne__(self._cast(other))

    def __repr__(self):
        return '%s(%s)' % (type(self).__name__, self.data.__repr__())

    __str__ = __repr__

    def __setitem__(self, key, value):
        self.mutated()
        return self.data.__setitem__(key, value)

    def clear(self):
        self.mutated()
        return self.data.clear()

    def copy(self):
        return type(self)(self)

    def get(self, *args):
        return self.data.get(*args)

    def items(self):
        return self.data.items()

    def keys(self):
        return self.data.keys()

    def pop(self, key):
        self.mutated()
        return self.data.pop(key)

    def popitem(self):
        self.mutated()
        return self.data.popitem()

    def setdefault(self, key, value):
        if key not in self.data:
            self.mutated()
        return self.data.setdefault(key, value)

    def update(self, mapping):
        self.mutated()
        return self.data.update(mapping)

    def values(self):
        return self.data.values()

    def mutated(self): #pragma no cover
        pass

    def _cast(self, other):
        if isinstance(other, DictWrapper):
            other = other.data
        return other


class ListWrapper(object):

    def __init__(self, *args):
        self.data = list(*args)

    def __contains__(self, index):
        return self.data.__contains__(index)

    def __delitem__(self, index):
        self.mutated()
        return self.data.__delitem__(index)

    def __delslice__(self, start, end):
        self.mutated()
        return self.data.__delslice__(start, end)

    def __eq__(self, other):
        return self.data.__eq__(other)

    def __getitem__(self, index):
        return self.data.__getitem__(index)

    def __getslice__(self, start, end):
        return self.data.__getslice__(start, end)

    def __hash__(self):
        return self.data.__hash__()

    def __iadd__(self, other):
        return type(self)(self.data.__iadd__(other))

    def __imul__(self, other):
        return type(self)(self.data.__imul__(other))

    def __iter__(self):
        return self.data.__iter__()

    def __len__(self):
        return self.data.__len__()

    def __mul__(self, other):
        return self.data.__mul__(other)

    def __ne__(self, other):
        return self.data.__ne__(other)

    def __repr__(self):
        return '%s(%s)' % (type(self).__name__, self.data.__repr__())

    __str__ = __repr__

    def __rmul__(self, other):
        return self.data.__rmul__(other)

    def __setitem__(self, index, value):
        self.mutated()
        return self.data.__setitem__(index, value)

    def __setslice__(self, start, end, seq):
        self.mutated()
        return self.data.__setslice__(start, end, seq)

    def append(self, value):
        self.mutated()
        return self.data.append(value)

    def count(self, value):
        return self.data.count(value)

    def extend(self, seq):
        self.mutated()
        return self.data.extend(seq)

    def index(self, value):
        return self.data.index(value)

    def insert(self, index, value):
        self.mutated()
        return self.data.insert(index, value)

    def pop(self, index):
        self.mutated()
        return self.data.pop(index)

    def remove(self, value):
        self.mutated()
        return self.data.remove(value)

    def reverse(self):
        self.mutated()
        return self.data.reverse()

    def sort(self, *args, **kw):
        self.mutated()
        return self.data.sort(*args, **kw)

    def mutated(self): #pragma no cover
        pass



import acidfs
import datetime
import json
import sys
import transaction

from .collection_wrappers import DictWrapper
from .collection_wrappers import ListWrapper

CHURRO_EXT = '.churro'
CHURRO_FOLDER = '__folder__' + CHURRO_EXT

if sys.version_info[0] == 2: # pragma NO COVER
    DECODE_MODE = 'rb'
    ENCODE_MODE = 'wb'
else:  # pragma NO COVER
    DECODE_MODE = 'r'
    ENCODE_MODE = 'w'


class Churro(object):
    """
    **Constructor Arguments**

    ``repo``

       The path to the repository in the real, local filesystem.

    ``head``

       The name of a branch to use as the head for this transaction.  Changes
       made using this instance will be merged to the given head.  The default,
       if omitted, is to use the repository's current head.

    ``factory``

       A callable that returns the root database object to be stored as the
       root when creating a new database.  The default factory returns an
       instance of `churro.PersistentFolder`.  This has no effect if the
       repository has already been created.

    ``create``

       If there is not a Git repository in the indicated directory, should one
       be created?  The default is `True`.

    ``bare``

       If the Git repository is to be created, create it as a bare repository.
       If the repository is already created or `create` is False, this argument
       has no effect.
    """
    session = None

    def __init__(self, repo, head='HEAD',
                 factory=None, create=True, bare=False):
        self.fs = acidfs.AcidFS(repo, head=head, create=create, bare=bare,
                                name='Churro.AcidFS')
        if factory is None:
            factory = PersistentFolder
        self.factory = factory

    def _session(self):
        """
        Make sure we're in a session.
        """
        if not self.session or self.session.closed:
            self.session = _Session(self.fs)
        return self.session

    def root(self):
        """
        Gets the root folder of the repository.  This is the starting point for
        traversing to other objects in the repository.
        """
        return self._session().get_root(self.factory)


_marker = object()
_removed = object()


class reify(object):
    # Stolen from Pyramid
    """ Put the result of a method which uses this (non-data)
    descriptor decorator in the instance dict after the first call,
    effectively replacing the decorator with an instance variable."""

    def __init__(self, wrapped):
        self.wrapped = wrapped
        try:
            self.__doc__ = wrapped.__doc__
        except: # pragma: no cover
            pass

    def __get__(self, inst, objtype=None):
        val = self.wrapped(inst)
        setattr(inst, self.wrapped.__name__, val)
        return val


class JsonCodec(object):
    """
    Encodes/decodes Python objects as JSON.
    """
    @staticmethod
    def encode_hook(obj):
        if not isinstance(obj, Persistent):
            return obj

        cls = type(obj)
        dotted_name = '%s.%s' % (cls.__module__, cls.__name__)
        data = {}
        for member in cls.mro():
            for name, prop in member.__dict__.items():
                if not isinstance(prop, PersistentProperty):
                    continue
                if name not in data:
                    data[name] = prop.to_json(prop.__get__(obj))
        return {
            '__churro_class__': dotted_name,
            '__churro_data__': data}

    def encode(self, obj, stream):
        json.dump(obj, stream, default=self.encode_hook, indent=4)

    @staticmethod
    def decode_hook(data):
        if '__churro_class__' not in data:
            return data

        cls = _resolve_dotted_name(data['__churro_class__'])
        obj = cls.__new__(cls)
        for name, value in data['__churro_data__'].items():
            prop = getattr(cls, name)
            prop.__set__(obj, prop.from_json(value), False)

        return obj

    def decode(self, stream):
        return json.load(stream, object_hook=self.decode_hook)


codec = JsonCodec()


class PersistentType(type):

    def __init__(cls, name, bases, members):
        type.__init__(cls, name, bases, members)
        for name, prop in members.items():
            if isinstance(prop, PersistentProperty):
                prop.set_name(name)


PersistentBase = PersistentType('PersistentBase', (object,), {})


class PersistentProperty(object):
    """
    The base type for all persistent properties.  This property type can handle
    any data type as a value that is serializable natively to JSON.  Other types
    are implemented by extending this class and overriding the `from_json`,
    `to_json`, and `validate` methods.
    """
    default = None

    def set_name(self, name):
        self.attr = '.' + name

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        return getattr(obj, self.attr, self.default)

    def __set__(self, obj, value, set_dirty=True):
        if set_dirty:
            obj.set_dirty()
        if isinstance(value, Persistent):
            value.__setinstance__(obj.__instance__)
        return setattr(obj, self.attr, self.validate(value))

    def from_json(self, value):
        """
        Converts a value from its JSON representation to a Python object.
        """
        return value

    def to_json(self, value):
        """
        Converts a value from a Python object to an object that can be
        serialized as JSON.
        """
        return value

    def validate(self, value):
        """
        Used at assignment time to validate a value.  If a value is not of the
        proper type and cannot be converted to the proper type, a `ValueError`
        is raised, otherwise the valua is returned, including any transformation
        or coercion that has been performed."""
        return value


class PersistentDate(PersistentProperty):
    """
    A persistent attribute type that can store instances of `datetime.date`.
    """

    def from_json(self, value):
        if value:
            return datetime.date(*map(int, value.split('-')))
        return value

    def to_json(self, value):
        if value:
            return '%s-%s-%s' % (value.year, value.month, value.day)
        return value

    def validate(self, value):
        if value is not None and not isinstance(value, datetime.date):
            raise ValueError("%s is not an instance of datetime.date")
        return value


class PersistentDatetime(PersistentProperty):
    """
    A persistent attribute type that can store instances of `datetime.datetime`.
    """

    def from_json(self, value):
        if value:
            d, t = value.split('T')
            year, month, day = map(int, d.split('-'))
            hour, minute, second = map(int, t.split(':'))
            return datetime.datetime(year, month, day, hour, minute, second)
        return value

    def to_json(self, value):
        if value:
            return '%04d-%02d-%02dT%02d:%02d:%02d' % (
                value.year, value.month, value.day, value.hour, value.minute,
                value.second)
        return value

    def validate(self, value):
        if value is not None and not isinstance(value, datetime.datetime):
            raise ValueError("%s is not an instance of datetime.datetime")
        return value


class Persistent(PersistentBase):
    """
    This is the base class from which all persistent classes for `Churro` must
    be derived.  Only objects which are instances of a class derived from
    `Persistent` may be stored in a `Churro` repository.
    """
    _dirty = True
    __name__ = None
    __parent__ = None
    _fs = None

    def __new__(cls, *args, **kw):
        obj = super(Persistent, cls).__new__(cls)
        obj.__instance__ = obj
        return obj

    def __setinstance__(self, instance):
        self.__instance__ = instance

    def set_dirty(self):
        """
        Calling this method alerts `Churro` that this object is `dirty` and
        should be persisted at commit time.  It is usually not necesary to call
        this method from application code, since `Churro` tries to detect
        object mutation whenever possible.  You may need to call this method
        from your application code, however, if you use mutable data structures
        that are not themselves `Persistent` as values of persistent properties,
        as `Churro` has no way of detecting mutations to those structures.
        """
        node = self.__instance__
        while node is not None:
            node._dirty = True
            node = getattr(node, '__parent__', None)


class PersistentFolder(Persistent):
    """
    Classes which derive from this class are not only persistent in `Churro` but
    have dict-like properties allowing them to contain children which are
    other persistent objects or folders.  Storing an instance of
    `PersistentFolder` in a `Churro` repository, creates a folder in the
    underlying filesystem, in which child objects are stored.  Instances of
    `PersistentFolder` are dict-like and are interacted with in the same way as
    standard Python dictionaries.
    """
    @reify
    def _contents(self):
        contents = {}
        fs = self._fs
        if fs is None:
            return contents
        path = resource_path(self)
        if not fs.exists(path):
            return contents
        with fs.cd(path):
            for fname in fs.listdir():
                if fname == CHURRO_FOLDER:
                    continue
                if fs.isdir(fname):
                    folder_data = '%s/%s' % (fname, CHURRO_FOLDER)
                    if fs.exists(folder_data):
                        contents[fname] = ('folder', None)
                elif fname.endswith(CHURRO_EXT):
                    contents[fname[:-7]] = ('object', None)
        return contents

    @property
    def _filtered_contents(self):
        return dict([(name, (type, obj)) for name, (type, obj)
                in self._contents.items() if obj is not _removed])

    def keys(self):
        """
        Returns the names of child objects.
        """
        return self._filtered_contents.keys()

    def values(self):
        """
        Returns an iterator over child objects.
        """
        for name, value in self.items():
            yield value

    def __iter__(self):
        """
        Returns an iterator over child names.
        """
        return iter(self._filtered_contents.keys())

    def items(self):
        """
        Returns an iterator over (child object's name, child object) tuples.
        """
        contents = self._filtered_contents
        for name, (type, obj) in contents.items():
            if obj is None:
                obj = self._load(name, type)
            yield name, obj

    def __len__(self):
        """
        Returns the number of children.
        """
        return len(self._filtered_contents)

    def __nonzero__(self):
        """
        Returns boolean indicating whether folder has any children.
        """
        return bool(self._filtered_contents)

    def __getitem__(self, name):
        """
        Retreives the child object of the given name.  Raises `KeyError` if the
        child is not found.
        """
        obj = self.get(name, _marker)
        if obj is _marker:
            raise KeyError(name)
        return obj

    def get(self, name, default=None):
        """
        Returns the child object of the given name.  Returns `default` if the
        child is not found.
        """
        contents = self._filtered_contents
        objref = contents.get(name)
        if not objref:
            return default
        type, obj = objref
        if obj is None:
            obj = self._load(name, type)
        return obj

    def __contains__(self, name):
        """
        Returns boolean indicating whether a child with the given name exists
        in the folder.
        """
        return name in self.keys()

    def __setitem__(self, name, other):
        """
        Adds a child to the folder with the given name.  If there is already
        another child with the same name in the folder, that child is
        overwritten.
        """
        type = 'folder' if isinstance(other, PersistentFolder) else 'object'
        self._contents[name] = (type, other)
        other.__parent__ = self
        other.__name__ = name
        other._fs = self._fs
        _set_dirty(other)

    def __delitem__(self, name):
        """
        Removes the child with the given name from  the folder.  Raises
        `KeyError` if there is no child with the given name.
        """
        if not self._remove(name):
            raise KeyError(name)

    remove = __delitem__

    def pop(self, name, default=_marker):
        objref = self._remove(name)
        if not objref:
            if default is _marker:
                raise KeyError(name)
            return default

        type, obj = objref
        if obj:
            return obj
        return self._load(name, type, False)

    def _remove(self, name):
        contents = self._contents
        objref = contents.get(name)
        if objref:
            type, obj = objref
            if obj is _removed:
                return None
            contents[name] = (type, _removed)
            _set_dirty(self)
        return objref

    def _load(self, name, type, cache=True):
        if type == 'folder':
            fspath = resource_path(self, name, CHURRO_FOLDER)
        else:
            fspath = resource_path(self, name) + CHURRO_EXT
        obj = codec.decode(self._fs.open(fspath, DECODE_MODE))
        obj.__parent__ = self
        obj.__name__ = name
        obj._fs = self._fs
        obj._dirty = False
        if cache:
            self._contents[name] = (type, obj)
        return obj

    def _save(self):
        fs = self._fs
        path = resource_path(self)
        if not fs.exists(path):
            fs.mkdir(path)
        for name, (type, obj) in self._contents.items():
            if obj is None:
                continue
            if type == 'folder':
                if obj is _removed:
                    fs.rmtree(resource_path(self, name))
                else:
                    obj._save()
            else:
                fspath = resource_path(self, name) + CHURRO_EXT
                if obj is _removed:
                    fs.rm(fspath)
                else:
                    codec.encode(obj, fs.open(fspath, ENCODE_MODE))
                    obj._dirty = False
        fspath = '%s/%s' % (path, CHURRO_FOLDER)
        codec.encode(self, fs.open(fspath, ENCODE_MODE))
        self._dirty = False

    #def __repr__(self):
    #    pass


class PersistentDict(DictWrapper, Persistent):
    """
    A `PersistentDict` is a Python `dict` work alike that marks its parent
    object as `dirty` whenever it is mutated, solving the problem of using
    mutable datastructures as values for persistent properties with `Churro`
    and eliminating the need to call :meth:`~churro.Persistent.set_dirty`
    in application code when updating the dictionary.
    """
    data = PersistentProperty()

    def mutated(self):
        self.set_dirty()

    def __setinstance__(self, instance):
        self.__instance__ = instance
        for obj in self.data.values():
            if isinstance(obj, Persistent):
                obj.__setinstance__(self.__instance__)

class PersistentList(ListWrapper, Persistent):
    """
    A `PersistentList` is a Python `dict` work alike that marks its parent
    object as `dirty` whenever it is mutated, solving the problem of using
    mutable datastructures as values for persistent properties with `Churro`
    and eliminating the need to call :meth:`~churro.Persistent.set_dirty`
    in application code when updating the list.
    """
    data = PersistentProperty()

    def mutated(self):
        self.set_dirty()

    def __setinstance__(self, instance):
        self.__instance__ = instance
        for obj in self.data:
            if isinstance(obj, Persistent):
                obj.__setinstance__(self.__instance__)


class _Session(object):
    closed = False
    root = None

    def __init__(self, fs):
        self.fs = fs
        transaction.get().join(self)

    def abort(self, tx):
        """
        Part of datamanager API.
        """
        self.close()

    tpc_abort = abort

    def tpc_begin(self, tx):
        """
        Part of datamanager API.
        """

    def commit(self, tx):
        """
        Part of datamanager API.
        """

    def tpc_vote(self, tx):
        """
        Part of datamanager API.
        """
        root = self.root
        if root is None or not root._dirty:
            # Nothing to do
            return

        root._save()

    def tpc_finish(self, tx):
        """
        Part of datamanager API.
        """

    def sortKey(self):
        return 'Churro'

    def close(self):
        self.closed = True

    def get_root(self, factory):
        if self.root is not None: # is not None
            return self.root

        fs = self.fs
        path = '/' + CHURRO_FOLDER
        if fs.exists(path):
            root = codec.decode(fs.open(path, DECODE_MODE))
            root._dirty = False
        else:
            root = factory()
        root._fs = fs
        root.__name__ = root.__parent__ = None
        self.root = root
        return root


def _resolve_dotted_name(name):
    names = name.split('.')
    path = names.pop(0)
    target = __import__(path)
    while names:
        segment = names.pop(0)
        path += '.' + segment
        try:
            target = getattr(target, segment)
        except AttributeError:
            __import__(path)
            target = getattr(target, segment)
    return target


def resource_path(obj, *elements):
    def _inner(obj, path):
        if obj.__parent__ is not None:
            _inner(obj.__parent__, path)
            path.append(obj.__name__)
        return path
    path = _inner(obj, [])
    if elements:
        path.extend(elements)
    return '/' + '/'.join(path)


def _set_dirty(obj):
    while obj is not None:
        obj._dirty = True
        obj = obj.__parent__



